@@ -14,6 +14,7 @@
 
 #include "access/genam.h"
 #include "access/heapam.h"
+#include "access/sysattr.h"
 #include "catalog/indexing.h"
 #include "catalog/pg_attrdef.h"
 #include "catalog/pg_type.h"
@@ -135,13 +136,18 @@ static void pg_decode_truncate(LogicalDecodingContext *ctx,
 #endif
 
 static void columns_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, bool addcomma, Oid reloid);
-static void tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool replident, bool addcomma, Oid reloid);
-static void pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool addcomma);
+static void tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, Bitmapset *bs, bool replident, bool addcomma, Oid reloid);
+static void pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, Bitmapset *bs, bool addcomma);
+static void identity_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, Bitmapset *bs);
 static bool parse_table_identifier(List *qualified_tables, char separator, List **select_tables);
 static bool string_to_SelectTable(char *rawstring, char separator, List **select_tables);
 static bool split_string_to_list(char *rawstring, char separator, List **sl);
 static bool split_string_to_oid_list(char *rawstring, char separator, List **sl);
 
+static bool pg_filter_by_action(int change_type, JsonAction actions);
+static bool pg_filter_by_table(List *filter_tables, char *schemaname, char *tablename);
+static bool pg_add_by_table(List *add_tables, char *schemaname, char *tablename);
+
 /* version 1 */
 static void pg_decode_begin_txn_v1(LogicalDecodingContext *ctx,
 					ReorderBufferTXN *txn);
@@ -284,9 +290,9 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 	}
 
 	/* pretty print */
-	strcpy(data->ht, "");
-	strcpy(data->nl, "");
-	strcpy(data->sp, "");
+	data->ht[0] = '\0';
+	data->nl[0] = '\0';
+	data->sp[0] = '\0';
 
 	/* add all tables in all schemas by default */
 	t = palloc0(sizeof(SelectTable));
@@ -489,9 +495,9 @@ pg_decode_startup(LogicalDecodingContext *ctx, OutputPluginOptions *opt, bool is
 
 			if (data->pretty_print)
 			{
-				strncpy(data->ht, "\t", 1);
-				strncpy(data->nl, "\n", 1);
-				strncpy(data->sp, " ", 1);
+				data->ht[0] = '\t';
+				data->nl[0] = '\n';
+				data->sp[0] = ' ';
 			}
 		}
 		else if (strcmp(elem->defname, "write-in-chunks") == 0)
@@ -870,14 +876,6 @@ pg_decode_commit_txn(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 {
 	JsonDecodingData *data = ctx->output_plugin_private;
 
-#if PG_VERSION_NUM >= 130000
-	if (rbtxn_has_catalog_changes(txn))
-#else
-	if (txn->has_catalog_changes)
-#endif
-		elog(DEBUG2, "txn has catalog changes: yes");
-	else
-		elog(DEBUG2, "txn has catalog changes: no");
 	elog(DEBUG2, "my change counter: " UINT64_FORMAT " ; # of changes: " UINT64_FORMAT " ; # of changes in memory: " UINT64_FORMAT, data->nr_changes, txn->nentries, txn->nentries_mem);
 	elog(DEBUG2, "# of subxacts: %d", txn->nsubtxns);
 
@@ -951,7 +949,7 @@ pg_decode_commit_txn_v2(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
  * replident: is this tuple a replica identity?
  */
 static void
-tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool replident, bool addcomma, Oid reloid)
+tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, Bitmapset *bs, bool replident, bool addcomma, Oid reloid)
 {
 	JsonDecodingData	*data;
 	int					natt;
@@ -1049,32 +1047,9 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 		if (attr->attisdropped || attr->attnum < 0)
 			continue;
 
-		/* Search indexed columns in whole heap tuple */
-		if (indexdesc != NULL)
-		{
-			int		j;
-			bool	found_col = false;
-
-			for (j = 0; j < indexdesc->natts; j++)
-			{
-				Form_pg_attribute	iattr;
-
-				/* See explanation a few lines above. */
-#if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
-				iattr = indexdesc->attrs[j];
-#else
-				iattr = TupleDescAttr(indexdesc, j);
-#endif
-
-				if (strcmp(NameStr(attr->attname), NameStr(iattr->attname)) == 0)
-					found_col = true;
-
-			}
-
-			/* Print only indexed columns */
-			if (!found_col)
-				continue;
-		}
+		/* Replica identity column? */
+		if (bs != NULL && !bms_is_member(attr->attnum - FirstLowInvalidHeapAttributeNumber, bs))
+			continue;
 
 		/* Get Datum from tuple */
 		origval = heap_getattr(tuple, natt + 1, tupdesc, &isnull);
@@ -1143,7 +1118,15 @@ tuple_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tu
 			}
 
 			appendStringInfo(&coltypes, "%s", comma);
-			escape_json(&coltypes, type_str);
+			/*
+			 * format_type() returns a quoted identifier, if
+			 * required. In this case, it doesn't need to enclose the type name
+			 * in double quotes.
+			 */
+			if (type_str[0] == '"')
+				appendStringInfo(&coltypes, "%s", type_str);
+			else
+				escape_json(&coltypes, type_str);
 
 			pfree(type_str);
 
@@ -1370,15 +1353,15 @@ columns_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple
 
 /* Print replica identity information */
 static void
-identity_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc)
+identity_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, Bitmapset *bs)
 {
 	/* Last parameter does not matter */
-	tuple_to_stringinfo(ctx, tupdesc, tuple, indexdesc, true, false, InvalidOid);
+	tuple_to_stringinfo(ctx, tupdesc, tuple, bs, true, false, InvalidOid);
 }
 
 /* Print primary key information */
 static void
-pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, TupleDesc indexdesc, bool addcomma)
+pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple, Bitmapset *bs, bool addcomma)
 {
 	JsonDecodingData	*data;
 	int					natt;
@@ -1389,10 +1372,6 @@ pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple
 
 	data = ctx->output_plugin_private;
 
-	/* no primary key */
-	if (indexdesc == NULL)
-		return;
-
 	initStringInfo(&pknames);
 	initStringInfo(&pktypes);
 
@@ -1422,31 +1401,9 @@ pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple
 		if (attr->attisdropped || attr->attnum < 0)
 			continue;
 
-		/* Search pk columns in whole heap tuple */
-		if (indexdesc != NULL)
-		{
-			int		j;
-			bool	found_col = false;
-
-			for (j = 0; j < indexdesc->natts; j++)
-			{
-				Form_pg_attribute	iattr;
-
-				/* See explanation a few lines above. */
-#if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
-				iattr = indexdesc->attrs[j];
-#else
-				iattr = TupleDescAttr(indexdesc, j);
-#endif
-
-				if (strcmp(NameStr(attr->attname), NameStr(iattr->attname)) == 0)
-					found_col = true;
-			}
-
-			/* Print only indexed columns */
-			if (!found_col)
-				continue;
-		}
+		/* Primary key column? */
+		if (bs != NULL && !bms_is_member(attr->attnum - FirstLowInvalidHeapAttributeNumber, bs))
+			continue;
 
 		typid = attr->atttypid;
 
@@ -1497,7 +1454,15 @@ pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple
 			}
 
 			appendStringInfo(&pktypes, "%s", comma);
-			escape_json(&pktypes, type_str);
+			/*
+			 * format_type() returns a quoted identifier, if
+			 * required. In this case, it doesn't need to enclose the type name
+			 * in double quotes.
+			 */
+			if (type_str[0] == '"')
+				appendStringInfo(&pktypes, "%s", type_str);
+			else
+				escape_json(&pktypes, type_str);
 
 			pfree(type_str);
 		}
@@ -1523,6 +1488,83 @@ pk_to_stringinfo(LogicalDecodingContext *ctx, TupleDesc tupdesc, HeapTuple tuple
 	pfree(pktypes.data);
 }
 
+static bool
+pg_filter_by_action(int change_type, JsonAction actions)
+{
+	if (change_type == REORDER_BUFFER_CHANGE_INSERT && !actions.insert)
+	{
+		elog(DEBUG3, "ignore INSERT");
+		return true;
+	}
+	if (change_type == REORDER_BUFFER_CHANGE_UPDATE && !actions.update)
+	{
+		elog(DEBUG3, "ignore UPDATE");
+		return true;
+	}
+	if (change_type == REORDER_BUFFER_CHANGE_DELETE && !actions.delete)
+	{
+		elog(DEBUG3, "ignore DELETE");
+		return true;
+	}
+
+	return false;
+}
+
+static bool
+pg_filter_by_table(List *filter_tables, char *schemaname, char *tablename)
+{
+	if (list_length(filter_tables) > 0)
+	{
+		ListCell	*lc;
+
+		foreach(lc, filter_tables)
+		{
+			SelectTable	*t = lfirst(lc);
+
+			if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
+			{
+				if (t->alltables || strcmp(t->tablename, tablename) == 0)
+				{
+					elog(DEBUG2, "\"%s\".\"%s\" was filtered out",
+								((t->allschemas) ? "*" : t->schemaname),
+								((t->alltables) ? "*" : t->tablename));
+					return true;
+				}
+			}
+		}
+	}
+
+	return false;
+}
+
+static bool
+pg_add_by_table(List *add_tables, char *schemaname, char *tablename)
+{
+	if (list_length(add_tables) > 0)
+	{
+		ListCell	*lc;
+
+		/* all tables in all schemas are added by default */
+		foreach(lc, add_tables)
+		{
+			SelectTable	*t = lfirst(lc);
+
+			if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
+			{
+				if (t->alltables || strcmp(t->tablename, tablename) == 0)
+				{
+					elog(DEBUG2, "\"%s\".\"%s\" was added",
+								((t->allschemas) ? "*" : t->schemaname),
+								((t->alltables) ? "*" : t->tablename));
+					return true;
+				}
+			}
+		}
+	}
+
+	return false;
+}
+
 /* Callback for individual changed tuples */
 static void
 pg_decode_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
@@ -1547,11 +1589,8 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	TupleDesc	tupdesc;
 	MemoryContext old;
 
-	Relation	indexrel;
-	TupleDesc	indexdesc;
-
-	Relation	pkrel = NULL;
-	TupleDesc	pkdesc = NULL;
+	Bitmapset	*pkbs = NULL;
+	Bitmapset	*ribs = NULL;
 
 	char		*schemaname;
 	char		*tablename;
@@ -1560,21 +1599,9 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 
 	data = ctx->output_plugin_private;
 
-	if (change->action == REORDER_BUFFER_CHANGE_INSERT && !data->actions.insert)
-	{
-		elog(DEBUG3, "ignore INSERT");
+	/* filter changes by action */
+	if (pg_filter_by_action(change->action, data->actions))
 		return;
-	}
-	if (change->action == REORDER_BUFFER_CHANGE_UPDATE && !data->actions.update)
-	{
-		elog(DEBUG3, "ignore UPDATE");
-		return;
-	}
-	if (change->action == REORDER_BUFFER_CHANGE_DELETE && !data->actions.delete)
-	{
-		elog(DEBUG3, "ignore DELETE");
-		return;
-	}
 
 	class_form = RelationGetForm(relation);
 	tupdesc = RelationGetDescr(relation);
@@ -1593,69 +1620,21 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	RelationGetIndexList(relation);
 
 	/* Filter tables, if available */
-	if (list_length(data->filter_tables) > 0)
+	if (pg_filter_by_table(data->filter_tables, schemaname, tablename))
 	{
-		ListCell	*lc;
-		bool		skip = false;
-
-		foreach(lc, data->filter_tables)
-		{
-			SelectTable	*t = lfirst(lc);
-
-			if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
-			{
-				if (t->alltables || strcmp(t->tablename, tablename) == 0)
-				{
-					elog(DEBUG2, "\"%s\".\"%s\" was filtered out",
-								((t->allschemas) ? "*" : t->schemaname),
-								((t->alltables) ? "*" : t->tablename));
-					skip = true;
-				}
-			}
-		}
-
-		/* table was found */
-		if (skip)
-		{
-			MemoryContextSwitchTo(old);
-			MemoryContextReset(data->context);
-			return;
-		}
+		MemoryContextSwitchTo(old);
+		MemoryContextReset(data->context);
+		return;
 	}
 
 	/* Add tables */
-	if (list_length(data->add_tables) > 0)
+	if (!pg_add_by_table(data->add_tables, schemaname, tablename))
 	{
-		ListCell	*lc;
-		bool		skip = true;
-
-		/* all tables in all schemas are added by default */
-		foreach(lc, data->add_tables)
-		{
-			SelectTable	*t = lfirst(lc);
-
-			if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
-			{
-				if (t->alltables || strcmp(t->tablename, tablename) == 0)
-				{
-					elog(DEBUG2, "\"%s\".\"%s\" was added",
-								((t->allschemas) ? "*" : t->schemaname),
-								((t->alltables) ? "*" : t->tablename));
-					skip = false;
-				}
-			}
-		}
-
-		/* table was not found */
-		if (skip)
-		{
-			MemoryContextSwitchTo(old);
-			MemoryContextReset(data->context);
-			return;
-		}
+		MemoryContextSwitchTo(old);
+		MemoryContextReset(data->context);
+		return;
 	}
 
-	/* Sanity checks */
 	switch (change->action)
 	{
 		case REORDER_BUFFER_CHANGE_INSERT:
@@ -1759,30 +1738,21 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	appendStringInfo(ctx->out, ",%s", data->nl);
 
 	if (data->include_pk)
-	{
 #if	PG_VERSION_NUM >= 100000
-		if (OidIsValid(relation->rd_pkindex))	/* 10+ */
-		{
-			pkrel = RelationIdGetRelation(relation->rd_pkindex);
-			pkdesc = RelationGetDescr(pkrel);
-		}
+		pkbs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_PRIMARY_KEY);
 #else
-		if (OidIsValid(relation->rd_replidindex) && relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
-		{
-			pkrel = RelationIdGetRelation(relation->rd_replidindex);
-			pkdesc = RelationGetDescr(pkrel);
-		}
+		pkbs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_KEY);
 #endif
-	}
 
 	switch (change->action)
 	{
 		case REORDER_BUFFER_CHANGE_INSERT:
 			/* Print the new tuple */
-			if (data->include_pk && pkrel != NULL)
+			if (data->include_pk && OidIsValid(relation->rd_replidindex) &&
+					relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
 			{
 				columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, true, change->data.tp.relnode.relNode);
-				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkdesc, false);
+				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkbs, false);
 			}
 			else
 			{
@@ -1792,8 +1762,9 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 		case REORDER_BUFFER_CHANGE_UPDATE:
 			/* Print the new tuple */
 			columns_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, true, change->data.tp.relnode.relNode);
-			if (data->include_pk && pkrel != NULL)
-				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkdesc, true);
+			if (data->include_pk && OidIsValid(relation->rd_replidindex) &&
+					relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
+				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkbs, true);
 
 			/*
 			 * The old tuple is available when:
@@ -1808,17 +1779,8 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 			{
 				elog(DEBUG1, "old tuple is null");
 
-				indexrel = RelationIdGetRelation(relation->rd_replidindex);
-				if (indexrel != NULL)
-				{
-					indexdesc = RelationGetDescr(indexrel);
-					identity_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, indexdesc);
-					RelationClose(indexrel);
-				}
-				else
-				{
-					identity_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, NULL);
-				}
+				ribs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_IDENTITY_KEY);
+				identity_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, ribs);
 			}
 			else
 			{
@@ -1827,21 +1789,12 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 			}
 			break;
 		case REORDER_BUFFER_CHANGE_DELETE:
-			if (data->include_pk && pkrel != NULL)
-				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkdesc, true);
+			if (data->include_pk && OidIsValid(relation->rd_replidindex) &&
+					relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
+				pk_to_stringinfo(ctx, tupdesc, &change->data.tp.newtuple->tuple, pkbs, true);
 
-			/* Print the replica identity */
-			indexrel = RelationIdGetRelation(relation->rd_replidindex);
-			if (indexrel != NULL)
-			{
-				indexdesc = RelationGetDescr(indexrel);
-				identity_to_stringinfo(ctx, tupdesc, &change->data.tp.oldtuple->tuple, indexdesc);
-				RelationClose(indexrel);
-			}
-			else
-			{
-				identity_to_stringinfo(ctx, tupdesc, &change->data.tp.oldtuple->tuple, NULL);
-			}
+			ribs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_IDENTITY_KEY);
+			identity_to_stringinfo(ctx, tupdesc, &change->data.tp.oldtuple->tuple, ribs);
 
 			if (change->data.tp.oldtuple == NULL)
 				elog(DEBUG1, "old tuple is null");
@@ -1852,8 +1805,8 @@ pg_decode_change_v1(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 			Assert(false);
 	}
 
-	if (data->include_pk && pkrel != NULL)
-		RelationClose(pkrel);
+	bms_free(pkbs);
+	bms_free(ribs);
 
 	appendStringInfo(ctx->out, "%s%s}", data->ht, data->ht);
 
@@ -1952,8 +1905,7 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 	JsonDecodingData	*data;
 	TupleDesc			tupdesc;
 	Relation			defrel = NULL;
-	Relation			idxrel;
-	TupleDesc			idxdesc = NULL;
+	Bitmapset			*bs = NULL;
 	int					i;
 	Datum				*values;
 	bool				*nulls;
@@ -1971,41 +1923,14 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 	/* figure out replica identity columns */
 	if (kind == PGOUTPUTJSON_IDENTITY)
 	{
-		if (OidIsValid(relation->rd_replidindex))		/* REPLICA IDENTITY INDEX */
-		{
-			idxrel = RelationIdGetRelation(relation->rd_replidindex);
-			idxdesc = RelationGetDescr(idxrel);
-		}
-#if	PG_VERSION_NUM >= 100000
-		else if (OidIsValid(relation->rd_pkindex))	/* REPLICA IDENTITY DEFAULT + PK (10+) */
-		{
-			idxrel = RelationIdGetRelation(relation->rd_pkindex);
-			idxdesc = RelationGetDescr(idxrel);
-		}
-#else
-		else if (relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT && OidIsValid(relation->rd_replidindex))	/* 9.4, 9.5 and 9.6 */
-		{
-			idxrel = RelationIdGetRelation(relation->rd_replidindex);
-			idxdesc = RelationGetDescr(idxrel);
-		}
-#endif
-		else if (relation->rd_rel->relreplident != REPLICA_IDENTITY_FULL)
-			elog(ERROR, "table does not have primary key or replica identity");
+		bs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_IDENTITY_KEY);
 	}
 	else if (kind == PGOUTPUTJSON_PK)
 	{
 #if	PG_VERSION_NUM >= 100000
-		if (OidIsValid(relation->rd_pkindex))	/* 10+ */
-		{
-			idxrel = RelationIdGetRelation(relation->rd_pkindex);
-			idxdesc = RelationGetDescr(idxrel);
-		}
+		bs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_PRIMARY_KEY);
 #else
-		if (OidIsValid(relation->rd_replidindex) && relation->rd_rel->relreplident == REPLICA_IDENTITY_DEFAULT)
-		{
-			idxrel = RelationIdGetRelation(relation->rd_replidindex);
-			idxdesc = RelationGetDescr(idxrel);
-		}
+		bs = RelationGetIndexAttrBitmap(relation, INDEX_ATTR_BITMAP_KEY);
 #endif
 	}
 
@@ -2022,9 +1947,6 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 	for (i = 0; i < tupdesc->natts; i++)
 	{
 		Form_pg_attribute	attr;
-		int					j;
-		bool				found = false;
-		char				*type_str;
 
 #if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
 		attr = tupdesc->attrs[i];
@@ -2036,40 +1958,13 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 		if (attr->attisdropped || attr->attnum < 0)
 			continue;
 
-		/*
-		 * oldtuple contains NULL on those values that are not defined by
-		 * REPLICA IDENTITY. In this case, print only non-null values.
-		 */
-		if (nulls[i] && (kind == PGOUTPUTJSON_PK || kind == PGOUTPUTJSON_IDENTITY))
+		if (bs != NULL && !bms_is_member(attr->attnum - FirstLowInvalidHeapAttributeNumber, bs))
 			continue;
 
 		/* don't send unchanged TOAST Datum */
 		if (!nulls[i] && attr->attlen == -1 && VARATT_IS_EXTERNAL_ONDISK(values[i]))
 			continue;
 
-		/*
-		 * Is it replica identity column? Print only those columns or all
-		 * columns if REPLICA IDENTITY FULL is set.
-		 */
-		if (kind == PGOUTPUTJSON_PK || (kind == PGOUTPUTJSON_IDENTITY && relation->rd_rel->relreplident != REPLICA_IDENTITY_FULL))
-		{
-			for (j = 0; j < idxdesc->natts; j++)
-			{
-				Form_pg_attribute	iattr;
-
-#if (PG_VERSION_NUM >= 90600 && PG_VERSION_NUM < 90605) || (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90509) || (PG_VERSION_NUM >= 90400 && PG_VERSION_NUM < 90414)
-				iattr = idxdesc->attrs[j];
-#else
-				iattr = TupleDescAttr(idxdesc, j);
-#endif
-				if (strcmp(NameStr(attr->attname), NameStr(iattr->attname)) == 0)
-					found = true;
-			}
-
-			if (!found)
-				continue;
-		}
-
 		if (need_sep)
 			appendStringInfoChar(ctx->out, ',');
 		need_sep = true;
@@ -2083,6 +1978,7 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 		{
 			HeapTuple		type_tuple;
 			Form_pg_type	type_form;
+			char			*type_str;
 
 			type_tuple = SearchSysCache1(TYPEOID, ObjectIdGetDatum(attr->atttypid));
 			type_form = (Form_pg_type) GETSTRUCT(type_tuple);
@@ -2097,7 +1993,15 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 				type_str = format_type_with_typemod(attr->atttypid, attr->atttypmod);
 
 			appendStringInfoString(ctx->out, ",\"type\":");
-			appendStringInfo(ctx->out, "\"%s\"", type_str);
+			/*
+			 * format_type_with_typemod() returns a quoted identifier, if
+			 * required. In this case, it doesn't need to enclose the type name
+			 * in double quotes.
+			 */
+			if (type_str[0] == '"')
+				appendStringInfo(ctx->out, "%s", type_str);
+			else
+				appendStringInfo(ctx->out, "\"%s\"", type_str);
 			pfree(type_str);
 
 			ReleaseSysCache(type_tuple);
@@ -2210,6 +2114,8 @@ pg_decode_write_tuple(LogicalDecodingContext *ctx, Relation relation, HeapTuple
 #endif
 	}
 
+	bms_free(bs);
+
 	pfree(values);
 	pfree(nulls);
 }
@@ -2222,7 +2128,6 @@ pg_decode_write_change(LogicalDecodingContext *ctx, ReorderBufferTXN *txn, Relat
 	/* make sure rd_pkindex and rd_replidindex are set */
 	RelationGetIndexList(relation);
 
-	/* sanity checks */
 	switch (change->action)
 	{
 		case REORDER_BUFFER_CHANGE_INSERT:
@@ -2405,21 +2310,9 @@ pg_decode_change_v2(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	char	*schemaname;
 	char	*tablename;
 
-	if (change->action == REORDER_BUFFER_CHANGE_INSERT && !data->actions.insert)
-	{
-		elog(DEBUG3, "ignore INSERT");
+	/* filter changes by action */
+	if (pg_filter_by_action(change->action, data->actions))
 		return;
-	}
-	if (change->action == REORDER_BUFFER_CHANGE_UPDATE && !data->actions.update)
-	{
-		elog(DEBUG3, "ignore UPDATE");
-		return;
-	}
-	if (change->action == REORDER_BUFFER_CHANGE_DELETE && !data->actions.delete)
-	{
-		elog(DEBUG3, "ignore DELETE");
-		return;
-	}
 
 	/* avoid leaking memory by using and resetting our own context */
 	old = MemoryContextSwitchTo(data->context);
@@ -2429,66 +2322,19 @@ pg_decode_change_v2(LogicalDecodingContext *ctx, ReorderBufferTXN *txn,
 	tablename = RelationGetRelationName(relation);
 
 	/* Exclude tables, if available */
-	if (list_length(data->filter_tables) > 0)
+	if (pg_filter_by_table(data->filter_tables, schemaname, tablename))
 	{
-		ListCell	*lc;
-		bool		skip = false;
-
-		foreach(lc, data->filter_tables)
-		{
-			SelectTable	*t = lfirst(lc);
-
-			if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
-			{
-				if (t->alltables || strcmp(t->tablename, tablename) == 0)
-				{
-					elog(DEBUG2, "\"%s\".\"%s\" was filtered out",
-								((t->allschemas) ? "*" : t->schemaname),
-								((t->alltables) ? "*" : t->tablename));
-					skip = true;
-				}
-			}
-		}
-
-		/* table was found */
-		if (skip)
-		{
-			MemoryContextSwitchTo(old);
-			MemoryContextReset(data->context);
-			return;
-		}
+		MemoryContextSwitchTo(old);
+		MemoryContextReset(data->context);
+		return;
 	}
 
 	/* Add tables */
-	if (list_length(data->add_tables) > 0)
+	if (!pg_add_by_table(data->add_tables, schemaname, tablename))
 	{
-		ListCell	*lc;
-		bool		skip = true;
-
-		/* all tables in all schemas are added by default */
-		foreach(lc, data->add_tables)
-		{
-			SelectTable	*t = lfirst(lc);
-
-			if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
-			{
-				if (t->alltables || strcmp(t->tablename, tablename) == 0)
-				{
-					elog(DEBUG2, "\"%s\".\"%s\" was added",
-								((t->allschemas) ? "*" : t->schemaname),
-								((t->alltables) ? "*" : t->tablename));
-					skip = false;
-				}
-			}
-		}
-
-		/* table was not found */
-		if (skip)
-		{
-			MemoryContextSwitchTo(old);
-			MemoryContextReset(data->context);
-			return;
-		}
+		MemoryContextSwitchTo(old);
+		MemoryContextReset(data->context);
+		return;
 	}
 
 	pg_decode_write_change(ctx, txn, relation, change);
@@ -2733,6 +2579,22 @@ static void pg_decode_truncate_v1(LogicalDecodingContext *ctx,
 	/* Avoid leaking memory by using and resetting our own context */
 	old = MemoryContextSwitchTo(data->context);
 
+	/* Exclude tables, if available */
+	if (pg_filter_by_table(data->filter_tables, schemaname, tablename))
+	{
+		MemoryContextSwitchTo(old);
+		MemoryContextReset(data->context);
+		continue;
+	}
+
+	/* Add tables */
+	if (!pg_add_by_table(data->add_tables, schemaname, tablename))
+	{
+		MemoryContextSwitchTo(old);
+		MemoryContextReset(data->context);
+		continue;
+	}
+
 	if (data->write_in_chunks)
 		OutputPluginPrepareWrite(ctx, true);
 
@@ -2819,53 +2681,19 @@ static void pg_decode_truncate_v2(LogicalDecodingContext *ctx,
 		tablename = RelationGetRelationName(relations[i]);
 
 		/* Exclude tables, if available */
-		if (list_length(data->filter_tables) > 0)
+		if (pg_filter_by_table(data->filter_tables, schemaname, tablename))
 		{
-			ListCell	*lc;
-
-			foreach(lc, data->filter_tables)
-			{
-				SelectTable	*t = lfirst(lc);
-
-				if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
-				{
-					if (t->alltables || strcmp(t->tablename, tablename) == 0)
-					{
-						elog(DEBUG2, "\"%s\".\"%s\" was filtered out",
-									((t->allschemas) ? "*" : t->schemaname),
-									((t->alltables) ? "*" : t->tablename));
-						continue;
-					}
-				}
-			}
+			MemoryContextSwitchTo(old);
+			MemoryContextReset(data->context);
+			continue;
 		}
 
 		/* Add tables */
-		if (list_length(data->add_tables) > 0)
+		if (!pg_add_by_table(data->add_tables, schemaname, tablename))
 		{
-			ListCell	*lc;
-			bool		skip = true;
-
-			/* all tables in all schemas are added by default */
-			foreach(lc, data->add_tables)
-			{
-				SelectTable	*t = lfirst(lc);
-
-				if (t->allschemas || strcmp(t->schemaname, schemaname) == 0)
-				{
-					if (t->alltables || strcmp(t->tablename, tablename) == 0)
-					{
-						elog(DEBUG2, "\"%s\".\"%s\" was added",
-									((t->allschemas) ? "*" : t->schemaname),
-									((t->alltables) ? "*" : t->tablename));
-						skip = false;
-					}
-				}
-			}
-
-			/* table was not found */
-			if (skip)
-				continue;
+			MemoryContextSwitchTo(old);
+			MemoryContextReset(data->context);
+			continue;
 		}
 
 		OutputPluginPrepareWrite(ctx, true);
